from DevOpsAPI import Wit


def test_create_get_delete(api, area):
    wi2 = api.WorkItem.create(Wit.Task, "this is a task", area=area)
    assert wi2.id > 0
    assert wi2.Title == "this is a task"
    wi = api.WorkItem.get(wi2.id)
    assert wi.id == wi2.id
    assert wi.Title == "this is a task"
    wi.delete()


def test_modify(api, area):
    wi = api.WorkItem.create(Wit.Task, "modify task", area=area)
    wi.Description = "description1"
    assert wi.Title == "modify task"
    assert wi.Description == "description1"
    wi.Title = "modified task title"
    wi.Description = "description2"
    wi2 = api.WorkItem.get(wi.id)
    assert wi2.Title == "modified task title"
    assert wi2.Description == "description2"
    wi.delete()


def test_find(api, area):
    ids = api.WorkItems.find({"System.WorkItemType": Wit.Task})
    for id in ids:
        api.WorkItem.get(id=id).delete()
    wi1 = api.WorkItem.create(Wit.Task, "list task 1", area=area)
    wi2 = api.WorkItem.create(Wit.Task, "list task 2", area=area)
    wi3 = api.WorkItem.create(Wit.Task, "list task 3", area=area)
    ids = api.WorkItems.find({"System.WorkItemType": Wit.Task})
    assert len(ids) == 3
    assert wi1.id in ids
    assert wi1.id in ids
    assert wi2.id in ids
    assert wi3.id in ids



from .connection import Connection
from .base import FunctionClass, FunctionManager
import json


class Project(FunctionClass):
    def __init__(self, _c, json=None):
        _c = Connection(_c.organization, None, _c.user, _c.apikey)
        super().__init__("projects", _c, json)


class Projects(FunctionManager):
    def __init__(self, _c):
        _c = Connection(_c.organization, None, _c.user, _c.apikey)
        super().__init__("projects", _class=Project, _c=_c)

    def get(self, id):
        return Project(self._c, json=self.get_properties(id))


def _dump(response):
    print(f"response: {json.dumps(response.json(), indent=4)}")



import json
import re
from lxml import etree


class Wit:
    Epic = "Epic"
    Issue = "Issue"
    Task = "Task"
    TestCase = "Test Case"
    Requirement = "Requirement"
    Epic = "Epic"
    Feature = "Feature"


class Step:
    def __init__(self, action, result=""):
        self.action = action
        self.result = result

    def as_xml(self):
        return Step._template.format(self.action, self.result)

    def __str__(self):
        return f"Step: action: {self.action}, result: {self.result}"

    _template = """
        <step id=\"2\" type=\"ValidateStep\">
            <parameterizedString isformatted=\"true\">{}</parameterizedString>
            <parameterizedString isformatted=\"true\">{}</parameterizedString><description/>
        </step>"""


class WorkItem:
    def __init__(self, connection, id=None, json=None):
        self._c = connection

        if id:
            self._id = id
        if json:
            self.json = json
            self._id = self.id

    @property
    def id(self):
        return self.json['id']

    @property
    def AssignedTo(self):
        return self.fields['System.AssignedTo']['displayName']

    @property
    def WorkItemType(self):
        return self.fields['System.WorkItemType']

    @property
    def Description(self):
        return self.fields["System.Description"]

    @Description.setter
    def Description(self, value):
        self._update_field("System.Description", value)

    @property
    def Title(self):
        return self.fields["System.Title"]

    @Title.setter
    def Title(self, value):
        self._update_field("System.Title", value)

    @property
    def Tags(self):
        return self.fields["System.Tags"]

    @Tags.setter
    def Tags(self, value):
        self._update_field("System.Tags", value)

    def _update_field(self, fieldname, value):
        patches = Patches()
        patches.append(Patch(Ops.replace, fieldname, value))
        response = self._c.patch(f"wit/workitems/{self.id}", json=patches.json(), is_json=True)
        if response.status_code != 200:
            print(json.dumps(response.json(), indent=4))
        response.raise_for_status()
        self.update()

    def update(self):
        # self._c.get("wit/fields")
        response = self._c.get(f"wit/workitems/{self._id}", expand="relations")
        self.json = response.json()
        return self

    @property
    def fields(self):
        return self.json['fields']

    @property
    def childs(self):
        ids = []
        for relation in self.json['relations']:
            if relation['rel'] != "System.LinkTypes.Hierarchy-Forward":
                continue
            m = re.match(r'https.*/(\d+)$', relation['url'])
            ids.append(int(m.group(1)))
        return ids

    def get(self, id):
        return WorkItem(self._c, id).update()

    def delete(self):
        response = self._c.delete(f"wit/workitems/{self.id}")
        response.raise_for_status()

    def create(self, type, title, area=None):
        patches = Patches()
        patches.append(Patch(Ops.add, "System.Title", title))
        if area:
            patches.append(Patch(Ops.add, "System.AreaPath", area))
        response = self._c.post(fnc=f"wit/workitems/${type}", json=patches.json(), is_json=True)
        response.raise_for_status()
        return WorkItem(self._c, None, response.json())

    def __str__(self):
        return f"{self.id}: {self.title} @{self.AssignedTo}"

    def __getattr__(self, name):
        _lookups = [f"System.{name}", f"Custom.{name}", name]
        for lookup in _lookups:
            if lookup in self.fields:
                return self.fields[lookup]
        return None


class TestCase(WorkItem):
    def create(self, title, area=None):
        wi = super().create(type=Wit.TestCase, title=title, area=area)
        return TestCase(wi._c, wi.id, wi.json)

    @property
    def Steps(self):
        steps = self.fields["Microsoft.VSTS.TCM.Steps"]
        xsteps = etree.fromstring(steps)
        return [Step(st[0].text, st[1].text) for st in xsteps]

    @Steps.setter
    def Steps(self, steps):
        strsteps = [step.as_xml() for step in steps]
        stepxml = f'<steps>{"".join(strsteps)}</steps>'

        patches = Patches()
        patches.append(Patch(Ops.add, "Microsoft.VSTS.TCM.Steps", stepxml))
        response = self._c.patch(f"wit/workitems/{self.id}", patches.json(), is_json=True)
        response.raise_for_status()
        self.update()

    def delete(self):
        response = self._c.delete(f"test/testcases/{self.id}")
        response.raise_for_status()

    def get(self, id):
        wi = WorkItem(self._c, id).update()
        return TestCase(wi._c, wi.id, wi.json)


class WorkItems:
    def __init__(self, connection):
        self._c = connection

    def find(self, filter={}):
        tokens = []
        for (attribute, value) in filter.items():
            if isinstance(value, dict):
                (op, value) = list(value.items())[0]
            elif value[0] == "~":
                value = value[1:]
                op = "Contains"
            else:
                op = "="
            tokens.append(f"[{attribute}] {op} '{value}'")
        tokenstr = " and ".join(tokens)
        ft = f"Where {tokenstr}" if len(tokens) > 0 else ""
        query = f"Select [System.Id] From WorkItems {ft}"
        search = {'query': query}
        response = self._c.post("wit/wiql", json=search)
        response.raise_for_status()
        return [wi['id'] for wi in response.json()['workItems']]


class Patch:
    def __init__(self, operation, field, value):
        self.operation = operation
        self.field = field
        self.value = value

    def json(self):
        result = {
            "from": None,
            "op": self.operation,
            "path": f"/fields/{self.field}",
            "value": self.value
        }
        return result


class Patches(list):
    def json(self):
        return [patch.json() for patch in self]


class Ops:
    add = "add"
    replace = "replace"



from enum import Enum
import json
import logging

from .base import FunctionManager, FunctionClass

log = logging.getLogger(__name__)


class TestResult:
    class State(Enum):
        Unspecified = 0
        NotStarted = 1
        InProgress = 2
        Completed = 3
        Aborted = 4
        Waiting = 5

    class Outcome:
        unspecified = "Unspecified"
        none = "None"
        passed = "Passed"
        failed = "Failed"
        inconclusive = "Inconclusive"
        timeout = "Timeout"
        aborted = "Aborted"
        blocked = "Blocked"
        notexecuted = "NotExecuted"
        warning = "warning"
        error = "error"
        notapplicable = "NotAppplicable"
        paused = "Paused"
        inprogress = "InProgress"
        notimpacted = "NotImpacted"

    def __init__(self, outcome, id, testpointid, testconfigurationid, title, rev):
        self.outcome = outcome
        self.id = id
        self.testpointid = testpointid
        self.testconfigurationid = testconfigurationid
        self.title = title
        self.rev = rev


class TestResults:
    def __init__(self, connection):
        self._c = connection

    def get(self, runid):
        response = self._c.get(f"test/Runs/{runid}/results?detailstoInclude=WorkItems")
        _dump(response)


class TestPoint(FunctionClass):
    def __init__(self, _c, json=None):
        return super().__init__("test/Plans", _c, json)

    @property
    def TestConfiguration(self):
        return TestConfigurations(self._c).get(self.configuration['id'])


class TestPoints(FunctionManager):
    def __init__(self, _c, planid, suiteid):
        super().__init__(f"test/Plans/{planid}/Suites/{suiteid}/points", TestPoint, _c, is_json=False)

    def find(self, tcid):
        query = f"?testCaseId={tcid}&isRecursive=true"
        response = self._c.get(self.fnc + query)
        return [TestPoint(self._c, json=js) for js in response.json()['value']]


class TestRun(FunctionClass):
    def __init__(self, _c, json=None):
        return super().__init__("test/runs", _c, json)

    def add(self, result):
        body = [{
            "testCase": {"id": result.id},
            "testPoint": {"id": result.testpointid},
            "testCaseTitle": result.title,
            "testCaseRevision": result.rev,
            "testCaseReferenceId": result.id,
            "configuration": {"id": result.testconfigurationid},
            # "TestCaseTitle": "huhu",
            # "automatedTestName": "huhu",
            # "id": result.id,
            # "project": "d676c54f-6ab2-4a78-9c44-43f0b7bd2d3d",
            "outcome": result.outcome
        }]

        logging.debug(body)
        response = self._c.post(f"test/runs/{self.id}/results", json=body, is_json=False)
        _dump(response)

    def close(self, state="Completed"):
        self.set('state', state)


class TestRuns(FunctionManager):
    def __init__(self, _c):
        super().__init__("test/runs", TestRun, _c, is_json=False)


class TestConfiguration(FunctionClass):
    def __init__(self, _c, json=None):
        return super().__init__("testplan/configurations", _c, json)

    @property
    def values(self):
        _result = {}
        for value in self.json['values']:
            _result[value['name']] = value['value']
        return _result


class TestConfigurations(FunctionManager):
    def __init__(self, _c):
        super().__init__("testplan/configurations", TestConfiguration, _c, is_json=False)


class TestSuiteTestCase(FunctionClass):
    def __init__(self, _c, json=None):
        super().__init__("test/Plans/{self.planid}/suites/{self.suiteid}/testcases", _c, json)
        # _dump2(json)

    @property
    def id(self):
        return int(self.testCase['id'])

    @property
    def configurations(self):
        _result = []
        pas = self.pointAssignments
        for pa in pas:
            pac = pa['configuration']
            _result.append(TestConfigurations(self._c).get(pac['id']))
        return _result


class TestSuiteTestCases(FunctionManager):
    def __init__(self, _c, planid, suiteid):
        self.planid = planid
        self.suiteid = suiteid
        super().__init__(f"test/Plans/{self.planid}/suites/{self.suiteid}/testcases", TestSuiteTestCase, _c, json)

    def add(self, ids):
        ids = [str(id) for id in ids]
        self._c.post(f"{self.fnc}/{','.join(ids)}")


class TestSuite(FunctionClass):
    def __init__(self, _c, json=None):
        self.planid = None
        super().__init__("test/Plans/{self.planid}/suites", _c, json)

    @property
    def TestCases(self):
        return TestSuiteTestCases(self._c, self.planid, self.id)

    @property
    def TestPoints(self):
        return TestPoints(self._c, planid=self.planid, suiteid=self.id)


# class TestPlanSuiteTestCases(FunctionManager):


class TestPlanSuites(FunctionManager):
    def __init__(self, _c, planid):
        self.planid = planid
        super().__init__(f"testplan/Plans/{planid}/suites", TestSuite, _c, is_json=False)

    def list(self):
        suites = super().list()
        for suite in suites:
            suite.planid = self.planid
        return suites


class TestPlan(FunctionClass):
    def __init__(self, _c, json):
        return super().__init__("testplan/plans", _c, json)

    @property
    def suites(self):
        return TestPlanSuites(self._c, planid=self.id).list()


class TestPlans(FunctionManager):
    def __init__(self, _c):
        super().__init__("testplan/plans", TestPlan, _c)


def _dump2(js):
    logging.debug(json.dumps(js, indent=4))


def _dump(response):
    log.debug(f"response: {json.dumps(response.json(), indent=4)}")



import json as js
import requests
import logging

log = logging.getLogger(__name__)


class Connection:
    def __init__(self, organization, project, user, apikey):
        self.organization = organization
        self.project = project
        self.user = user
        self.apikey = apikey
        self.headers = {'Content-type': 'application/json-patch+json'}

    def get(self, *args, **kwargs):
        return self.request("get", *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self.request("delete", *args, **kwargs)

    def post(self, *args, **kwargs):
        return self.request("post", *args, **kwargs)

    def patch(self, *args, **kwargs):
        return self.request("patch", *args, **kwargs)

    def request(self, rt, fnc, expand=None, json=None, is_json=False):
        headers = self.headers if is_json else None
        # log.warning(f"func is {fnc}")
        log.debug(f"calling: {rt}:{self._uri(fnc, expand)}")
        response = requests.request(rt, self._uri(fnc, expand), json=json, auth=self._auth, headers=headers)
        if response.status_code >= 400:
            response.raise_for_status()
        if response.status_code >= 300:
            log.error(js.dumps(response.json(), indent=4))
            logging.warning(js.dumps(response.json(), indent=4))
            response.raise_for_status()
        return response

    @property
    def _auth(self):
        return (self.user, self.apikey)

    def _uri(self, fnc, expand):
        project = f"/{self.project}" if self.project else ""
        expand = f"$expand={expand}&" if expand else ""
        return f"https://dev.azure.com/{self.organization}{project}/_apis/{fnc}?{expand}api-version=7.0"



import logging
import json

log = logging.getLogger(__name__)


class FunctionClass:
    def __init__(self, fnc, _c, json=None):
        self.fnc = fnc
        self._c = _c
        self.json = json

    def delete(self):
        self._c.delete(f"{self.fnc}/{self.id}")

    def __getattr__(self, name):
        return self.json[name]

    def set(self, name, value):
        update = {name: value}
        self._c.patch(f"{self.fnc}/{self.id}", json=update)


class FunctionManager:
    def __init__(self, fnc, _class, _c, is_json=False):
        self._class = _class
        self._c = _c
        self.fnc = fnc
        self.is_json = is_json

    def list(self):
        return self._list(self._class)

    def get(self, id):
        return self._class(self._c, json=self.get_properties(id))

    def create(self, **kwargs):
        return self._class(self._c, self._create(kwargs))

    def _list(self, classdef=None):
        response = self._c.get(self.fnc)
        if classdef:
            return [classdef(self._c, json=js) for js in response.json()['value']]
        else:
            return response.json()['values']

    def get_properties(self, id):
        url = f"/{self.fnc}/{id}"
        response = self._c.get(url)
#        _dump(response)
        return response.json()

    def _create(self, data):
        response = self._c.post(self.fnc, json=data, is_json=self.is_json)
        log.debug(f"FunctionManager::create: response: {_dump(response)}")
        return response.json()


def _dump(response):
    log.debug(f"response: {json.dumps(response.json(), indent=4)}")



from .wi import WorkItem, WorkItems, TestCase
from .testing import TestRun, TestResults, TestRuns, TestPlans, TestConfiguration, TestConfigurations
from .testing import TestPoint, TestPoints
from .wi import Wit, Step  # noqa: F401
from .projects import Projects, Project  # noqa: F401
from .connection import Connection


class Api:
    def __init__(self, organization, project, user, apikey):
        self._connection = Connection(organization, project, user, apikey)

    @property
    def WorkItem(self):
        return WorkItem(self._connection)

    @property
    def TestCase(self):
        return TestCase(self._connection)

    @property
    def TestCases(self):
        return TestCase(self._connection)

    @property
    def WorkItems(self):
        return WorkItems(self._connection)

    @property
    def TestRun(self):
        return TestRun(self._connection)

    def TestPoints(self, planid, suiteid):
        return TestPoints(self._connection, planid, suiteid)

    @property
    def TestRuns(self):
        return TestRuns(self._connection)

    @property
    def TestResults(self):
        return TestResults(self._connection)

    @property
    def TestPlans(self):
        return TestPlans(self._connection)

    @property
    def TestConfiguration(self):
        return TestConfiguration(self._connection)

    @property
    def TestConfigurations(self):
        return TestConfigurations(self._connection)

    @property
    def Projects(self):
        return Projects(self._connection)



from DevOpsAPI import Wit, Step


def test_create_get_delete(api, area):
    tc1 = api.TestCase.create("this is a Test Case", area=area)
    assert tc1.id > 0
    assert tc1.Title == "this is a Test Case"
    assert tc1.WorkItemType == Wit.TestCase
    tc11 = api.TestCase.get(tc1.id)
    assert tc11.id == tc1.id
    assert tc11.Title == "this is a Test Case"
    wi = api.WorkItem.get(tc1.id)
    assert wi.WorkItemType == Wit.TestCase
    tc1.delete()


def test_steps(api, area):
    tc = api.TestCase.create("Test Case with steps", area=area)
    tc.steps = [Step("in1", "out1"), Step("in2", "out2")]
    steps = tc.steps
    assert steps[0].action == "in1"
    assert steps[0].result == "out1"
    assert steps[1].action == "in2"
    assert steps[1].result == "out2"


def test_modify(api, area):
    tc = api.TestCase.create("Another Test Case", area=area)
    assert tc.Title == "Another Test Case"
    assert tc.WorkItemType == Wit.TestCase
    tc.Title = "Changed Title of Test Case"
    assert tc.Title == "Changed Title of Test Case"
    tc.delete()


def test_find(api, area):
    ids = api.WorkItems.find({"System.WorkItemType": Wit.TestCase})
    for id in ids:
        api.TestCase.get(id=id).delete()
    wi1 = api.TestCase.create("TC 1", area=area)
    wi1.Description = "This is a test case"
    wi2 = api.TestCase.create("TC 2", area=area)
    wi3 = api.TestCase.create("TC 3", area=area)
    ids = api.WorkItems.find({"System.WorkItemType": Wit.TestCase})
    assert len(ids) == 3
    assert wi1.id in ids
    assert wi2.id in ids
    wi11 = api.TestCase.get(wi1.id)
    assert wi11.Description == "This is a test case"

    assert wi3.id in ids



import pytest
import json

from DevOpsAPI import TestResult


@pytest.fixture
def testplan(api):
    tp = api.TestPlans.create(name="pytest_testplan")
    yield tp
    tp.delete()


@pytest.fixture
def testrun(api):
    tr = api.TestRuns.create(name="pytest_testrun")
    yield tr
    tr.delete()


@pytest.fixture
def testcase(api, area):
    tc = api.TestCase.create("this is a Test Case", area=area)
    yield tc
    tc.delete()


def test_get_testrun(testrun, api):
    tr2 = api.TestRuns.get(testrun.id)
    assert tr2.id == testrun.id
    assert tr2.name == testrun.name


def disabled_test_delete_all_testplans(api):
    tpl = api.TestPlans.list()
    for tp in tpl:
        tp.delete()


def test_create_list_delete_testplan(api):
    tp = api.TestPlans.create(name="automated testplan")
    assert tp.name == "automated testplan"
    assert tp.id > 0
    tpl = api.TestPlans.list()
    assert tp.name in [tp.name for tp in tpl]
    tp.delete()
    tpl = api.TestPlans.list()
    assert tp.name not in [tp.name for tp in tpl]


def test_create_list_delete_testruns(api):
    tr1 = api.TestRuns.create(name="tmp_testrun")
    tr2 = api.TestRuns.create(name="tmp_testrun")
    tr3 = api.TestRuns.create(name="tmp_testrun")
    trs = api.TestRuns.list()
    for tr in (tr1, tr2, tr3):
        assert tr.name in [ltr.name for ltr in trs]
    trs = api.TestRuns.list()
    for tr in trs:
        if tr.name == "tmp_testrun" and tr.state != "255":
            tr.delete()
    trs = api.TestRuns.list()
    for tr in trs:
        assert "tmp_testrun" not in tr.name or tr.state == "255"


def test_add_testrun_to_testplan(testplan, testcase, area, api):
    tp = api.TestPlans.create(name="pytest_addtestrun")
    ts = tp.suites[0]
    assert ts.name == "pytest_addtestrun"
    assert len(ts.TestCases.list()) == 0
    ts.TestCases.add([testcase.id])
    tcs = ts.TestCases.list()
    assert len(tcs) == 1
    assert tcs[0].id == testcase.id
    # tc1 = api.TestCase.create("pytest_addtestrun_testcase1", area=area)
    # tp.add_testcase(tc1)
    # tr = api.TestRuns.create(name="testrun1", plan={'id': tp.id})
    # assert tr


def test_create_testrun(api):
    tr = api.TestRuns.create(name="create_testrun")
    assert tr.id > 0
    assert tr.name == "create_testrun"
    assert not tr.isAutomated
    tr.delete()


def disabled_test_add_testresults(testcase, api, area):
    wi2 = api.TestCase.create("Another testcase", area=area)
    _dump2(wi2.json)
    tr = api.TestRuns.create(name="add_testcases", automated=False)
    assert tr.id > 0
    assert tr.name == "add_testcases"
    assert not tr.isAutomated
    r1 = TestResult(TestResult.Outcome.passed, str(testcase.id))
    assert tr.add(r1)


def _dump2(js):
    print(f"response: {json.dumps(js, indent=4)}")
